import 'dart:io';
import 'dart:convert' as c;

class FetchConfig {
  static String scheme = 'http://';
  static String host = 'localhost';
  static int port = 3000;
}

class Signal {
  late Function() _onSignal;
  Signal();
}

class AbortController {
  final signal = new Signal();

  abort() {
    signal._onSignal.call();
  }
}

class AbortError extends Error {}

class Response {
  final HttpClientRequest _req;
  final HttpClientResponse _res;

  const Response(this._req, this._res);

  Future<String> text() async {
    return await c.utf8.decodeStream(_res);
  }

  Future<dynamic> json() async {
    return c.json.decode(await text());
  }
}

final _client = HttpClient();

Future<Response> fetch(
  String url, {
  String method = 'GET',
  dynamic? body,
  Map<String, String>? headers,
  Signal? signal,
}) async {
  var uri = Uri.parse(url);

  if (uri.scheme.isEmpty) {
    uri = Uri(
      scheme: FetchConfig.scheme,
      host: FetchConfig.host,
      port: FetchConfig.port,
      path: uri.path,
      query: uri.query,
    );
  }

  final req = await _client.open(method, uri.host, uri.port, uri.path);
  signal?._onSignal = () => req.abort(AbortError(), StackTrace.current);

  if (headers != null) {
    headers.forEach((key, value) {
      req.headers.add(key, value);
    });
  }

  if (body is String) {
    req.write(body);
  } else if (body is List<int>) {
    req.add(body);
  }

  final res = await req.close();

  return Response(req, res);
}
